/* 
 * Exemple de selection : Arretez les spheres en cliquant dessus. Pour
 * tricher, appuyez sur espace ou cliquez et bougez sans relacher le
 * bouton.
 *
 * Auteur : Fares BELHADJ (Lire COPYING)
 * Email  : amsi@ai.univ-paris8.fr
 * Date   : 14/01/2008
 */

#include <math.h>
#include <stdio.h>
#include <assert.h>
#include <stdlib.h>
#include <SDL.h>
#include <GL/gl.h>
#include <GL/glu.h>

/*
 * Infos sur les spheres ...
 */
typedef struct objet_t objet_t;

struct objet_t {
  GLuint id; /* identifiant (nom) utilise dans le mode selection */
  GLfloat p[3], r, a, da; /* position, rayon, angle et variatioon d'angle */
  GLfloat color[4]; /* material */
};

/*
 * Pour imprimer les infos de debug mettre if(1) a la place de if(0)
 */
#define DEBUG if(0) fprintf

#define MIN(a, b) (((a) < (b)) ? (a) : (b))
#define MAX(a, b) (((a) > (b)) ? (a) : (b))

#define BUFSIZE 512
#define INF 1000000

/* Calcule un nombre pseudo-aleatoire dans l'intervalle [0, 1[. Ici la
 * distribution est uniforme.
 */
#define u_rand() (rand() / (RAND_MAX + 1.0))
/* Calcule un nombre pseudo-aleatoire dans l'intervalle [-1, 1[. Ici
 * la distribution est uniforme.
 */
#define su_rand() (2.0 * (rand() / (RAND_MAX + 1.0)) - 1.0)

/*
 * Prototypes des fonctions statiques.
 */
static SDL_Surface * sdlInit        (int w, int h);
static void          sdlInitGL      (SDL_Surface * s);
static void          sdlResizeGL    (SDL_Surface * s);
static void          sdlLoop        (SDL_Surface * s);
static int           sdlManageEvents(SDL_Surface * s);
static void          sdlDraw        (SDL_Surface * s);
static void          scene          (void);
static void          processHits    (GLint hits, GLuint * ptr);
static void          initObjets     (int n);
static void          initSphere     (int slices, int stacks);

/* 
 * Le tableau de spheres, leur nombre et LA liste d'affichage
 * correspondante.
 */
static objet_t * objets = NULL;
static GLuint n_objets = 0, sphereId = 0;

/*
 * pause ? et coordonnes de la souris.
 */
static int pause = 0, drag_x = -INF, drag_y = -INF;

/* un tout petit main */
int main(int argc, char ** argv) {
  SDL_Surface * s;
  s = sdlInit(640, 480);
  sdlLoop(s);
  return 0;
}

/* sdlInit : initialiser SDL pour OpenGL */
static SDL_Surface * sdlInit(int w, int h) {
  SDL_Surface * s;
  const SDL_VideoInfo * vi = NULL;
  if ( SDL_Init(SDL_INIT_VIDEO) < 0 ) {
    fprintf(stderr, "Impossible d'initialiser SDL: %s\n", SDL_GetError());
    exit(1);
  }

  if( (vi = SDL_GetVideoInfo()) == 0 ) {
    fprintf(stderr, "Echec de SDL_GetVideoInfo(): %s\n", SDL_GetError());
    exit(1);
  }
  SDL_GL_SetAttribute( SDL_GL_RED_SIZE, 5 );
  SDL_GL_SetAttribute( SDL_GL_GREEN_SIZE, 5 );
  SDL_GL_SetAttribute( SDL_GL_BLUE_SIZE, 5 );
  SDL_GL_SetAttribute( SDL_GL_DEPTH_SIZE, 16 );
  SDL_GL_SetAttribute( SDL_GL_DOUBLEBUFFER, 1 );
  if ( (s = SDL_SetVideoMode(w, h, vi->vfmt->BitsPerPixel,
			     SDL_OPENGL | SDL_RESIZABLE)) == NULL ) {
    fprintf(stderr, "Impossible d'ouvrir le mode video : %s\n", SDL_GetError());
    exit(1);
  }
  atexit(SDL_Quit);
  sdlInitGL(s);
  return s;
}

/* sdlInitGL : configurer les options GL */
/* et creer la liste d'affichage pour le plan */
static void sdlInitGL(SDL_Surface * s) {
  glShadeModel(GL_SMOOTH);
  glDepthFunc(GL_LEQUAL);
  glFrontFace(GL_CCW);
  glCullFace(GL_BACK);
  glClearColor(0, 0, 0, 0);
  glClearDepth(1.0);
  glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);
  glDisable(GL_CULL_FACE);
  glEnable(GL_DEPTH_TEST);

  glPolygonMode(GL_FRONT, GL_FILL);
  glPolygonMode(GL_BACK, GL_LINE);

  glEnable(GL_LIGHTING);
  glEnable(GL_LIGHT0);
  
  glDisable(GL_TEXTURE_2D);

  initObjets(40);
  initSphere(20, 20);
  
  sdlResizeGL(s);

}

/* 
 * CallBack : en cas de resize modifier la matrice de projection en
 * fonction du rapport w/h
 */
static void sdlResizeGL(SDL_Surface * s) {
  glViewport(0, 0, s->w, s->h);
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  gluPerspective(45.0, s->w / (float)s->h, 0.1, 100);
  glMatrixMode( GL_MODELVIEW );
  glLoadIdentity( );
}

/*
 * loop principale.
 *
 * On recupere le temps passe d'une frame a l'autre ; si pause est
 * faux, faire tourner le objets ; si la souris est cliquee, passer en
 * mode selection, rendre la scene, recuperer le nombre d'elements (en
 * meme temps repasser en mode rendu) dans la selection et les arreter
 * (dans processHits).
 */
static void sdlLoop(SDL_Surface * s) {
  Uint32 dt;
  GLint hits;
  Uint32 t0 = 0;
  int viewport[4];
  GLuint selectBuf[BUFSIZE];

  for(;;) {
    dt = SDL_GetTicks() - t0;
    t0 += dt;
    while(sdlManageEvents(s)) SDL_Delay(1);

    if(!pause) {
      int i, encore = 0;
      for(i = 0; i < n_objets; i++)
	if(objets[i].da != 0.0) {
	  encore = 1;
	  objets[i].a += objets[i].da * (dt / 1000.0);
	}
      if(!encore) initObjets(n_objets);
    }
    
    if(drag_x != -INF && drag_y != - INF) {
      glGetIntegerv(GL_VIEWPORT, viewport); /* recuperer le viewport en cours */
      glSelectBuffer(BUFSIZE, selectBuf); /* initialiser le buffer de selection pour GL */
      glRenderMode(GL_SELECT); /* passer en mode selection */
      
      glMatrixMode(GL_PROJECTION); /* charger le mode projection */
      glPushMatrix(); /* sauvegarder la matrice */
      glLoadIdentity(); /* charger la matrice identite */
      gluPickMatrix(drag_x, viewport[3] - drag_y, 5, 5, viewport); /* modifier la projection pour restreindre la scene a 2 pixxels autour des coordonnees de la souris */
      gluPerspective(45.0, viewport[2] / (float)viewport[3], 0.1, 100.0);
      glMatrixMode( GL_MODELVIEW ); /* repasser en modelview */
      scene(); /* dessiner la scene en mode selection */
      glMatrixMode(GL_PROJECTION); /* repasser en mode projection */
      glPopMatrix(); /* revenir a la projection initiale */
      glMatrixMode( GL_MODELVIEW );  /* repasser en modelview */
      
      hits = glRenderMode(GL_RENDER);  /* repasser en mode rendu et recuperer le nombre d'elements (noms) presents dans le precedent mode (selection) */
      processHits(hits, selectBuf); /* traiter la liste obtenue par le mode selection */
    }
    sdlDraw(s); /* appelera la scene en mode rendu */
  }
}

/* 
 * sdlManageEvents gere les evenements ; aussi renvoie vrai si la
 * fenetre est inactive, sinon faux.
 * Les evenements sont : 
 *
 * Les fleches pour la gestion de la camera
 *
 * escape, q, Q : quitter
 *
 * p : pause
 *
 * w : activer / desactiver le mode wireframe
 *
 * l : activer / desactiver l'eclairage
 *
 * f : activer / desactiver le dgrad
 *
 * F1 : fullscreen On / Off
 * 
 * redimensionnement de fenetre : appeler sdlResizeGL
 *
 */
static int sdlManageEvents(SDL_Surface * s) {
  GLint v[2];
  static int active = 1;
  SDL_Event event;
  while(SDL_PollEvent(&event)) 
    switch (event.type) {
    case SDL_MOUSEBUTTONDOWN:
      drag_x = event.button.x;
      drag_y = event.button.y;
      break;
    case SDL_MOUSEBUTTONUP:
      drag_x = drag_y = -INF;
      break;
    case SDL_MOUSEMOTION:
      if(drag_x != -INF && drag_y != -INF) {
	drag_x = event.motion.x;
	drag_y = event.motion.y;
      }
      break;
    case SDL_KEYDOWN:
      switch(event.key.keysym.sym) {
      case ' ':
      case 'p':
	pause = !pause;
	break;
      case 'w':
	glGetIntegerv(GL_POLYGON_MODE, v);
	if(v[0] == GL_FILL)
	  glPolygonMode(GL_FRONT, GL_LINE);
	else
	  glPolygonMode(GL_FRONT, GL_FILL);
	break;
      case 'f':
	glGetIntegerv(GL_SHADE_MODEL, v);
	if(v[0] == GL_FLAT)
	  glShadeModel(GL_SMOOTH);
	else
	  glShadeModel(GL_FLAT);
	break;
      case 'l':
	if(glIsEnabled(GL_LIGHTING)) {
	  glDisable(GL_LIGHTING);
	  glDisable(GL_LIGHT0);
	} else {
	  glEnable(GL_LIGHTING);
	  glEnable(GL_LIGHT0);
	}
	break;
      case SDLK_ESCAPE:
      case 'q':
	exit(0);
      case SDLK_F1:
	SDL_WM_ToggleFullScreen(s);
	break;
      default:
	fprintf(stderr, "La touche %s a ete pressee\n",
		SDL_GetKeyName(event.key.keysym.sym));
	break;
      }
      break;
    case SDL_KEYUP:
      break;
    case SDL_ACTIVEEVENT:
      if(event.active.state & SDL_APPACTIVE)
	active = event.active.gain;
      break;
    case SDL_VIDEORESIZE:
      if( (s = SDL_SetVideoMode(event.resize.w, event.resize.h, 
				s->format->BitsPerPixel, s->flags)) == NULL ) {
	fprintf(stderr, "Echec de SDL_SetVideoMode apres le resize : %s\n", 
		SDL_GetError());
	exit(1);
      }
      sdlResizeGL(s);
      break;
    case SDL_QUIT:
      exit(0);
    }
  return !active;
}

/*
 * dessiner la scene : n'est appelee qu'en mode rendu
 */
static void sdlDraw(SDL_Surface * s) {
  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
  scene();
  SDL_GL_SwapBuffers();
}

/*
 * La scene composee d'objets spheres tournant autour de l'axe y. Pour
 * chaque objet nous avons cree un nom (id) pour pouvoir le recuperer,
 * si besoin, en mode selection.
 */
static void scene(void) {
  int i;
  glPushMatrix(); 
  glInitNames();
  glPushName(0);
  glLoadIdentity();
  glTranslatef(0.0f,0.0f,-3.5f);
  for(i = 0; i < n_objets; i++) {
    glPushMatrix();
    glLoadName(objets[i].id);
    glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, objets[i].color);
    glRotatef(objets[i].a, 0, 1, 0);
    glTranslatef(objets[i].p[0], objets[i].p[1], objets[i].p[2]);
    glScalef(objets[i].r, objets[i].r, objets[i].r);
    glCallList(sphereId);
    glPopMatrix();
  }
  glPopMatrix();   
}

/*
 * Traitement du resultat obtenu dans le buffer de selection a la
 * sortie du mode selection. Cette fonction a ete creee a partir de la
 * fonction exemple processHits du redbook. Les objets trouves sont
 * stoppes : leur variation d'angle (da) passe a 0.
 */
static void processHits(GLint hits, GLuint * ptr) {
  int i, j;
  GLuint names;
  if(hits)
    DEBUG(stderr, "hits = %d\n", hits);
  for(i = 0; i < hits; i++) {
    names = *ptr;
    DEBUG(stderr, "number of names for hit_%d = %d\n", i, names); ptr++;
    DEBUG(stderr, "\tz1 = %f;\n", (float) *ptr/((unsigned int)-1)); ptr++;
    DEBUG(stderr, "\tz2 = %f;\n", (float) *ptr/((unsigned int)-1)); ptr++;
    DEBUG(stderr, "\tthe name is ");
    for(j = 0; j < names; j++) {
      objets[*ptr - 1].da = 0;
      DEBUG(stderr, "%d ", *ptr);
      ptr++;
    }
    DEBUG(stderr, "\n");
  }
}

/*
 * Initialiser la liste (tableau) des objets a n objets et les
 * parametrer.
 */
static void initObjets(int n) {
  float s;
  int i, j;
  if(objets) {
    free(objets);
    objets = NULL;
  }
  objets = malloc((n_objets = n) * sizeof * objets);
  assert(objets);
  for(i = 0; i < n_objets; i++) {
    objets[i].id = i + 1;
    objets[i].r  = 0.05 + 0.1 * u_rand();
    objets[i].a  = 0.0;
    objets[i].da = 45.0 + 90.0 * u_rand();
    for(j = 0; j < 3; j++)
      objets[i].p[j] = su_rand();
    do {
      s = 0;
      for(j = 0; j < 3; j++)
        s += (objets[i].color[j] = MAX(0, su_rand()));
    } while(s < 0.4);
    objets[i].color[4] = 1.0;
  }
}

/*
 * Initialiser la liste d'affichage correspondant  une sphere.
 */
static void initSphere(int slices, int stacks) {
  GLUquadricObj * q;
  q = gluNewQuadric();
  assert(q);
  gluQuadricNormals(q, GLU_SMOOTH); /* smooth est par defaut */
  gluQuadricTexture(q, GL_FALSE); /* false est par defaut */
  if(sphereId) {
    glDeleteLists(sphereId, 1);
  }
  sphereId = glGenLists(1);
  glNewList( sphereId, GL_COMPILE );
  gluSphere(q, 1.0, slices, stacks);
  glEndList();

  gluDeleteQuadric(q);
}
